/*************************************************
* Mardi.cpp
* Copyright (c) Rennie deGraaf, 2005.  All rights reserved.
* $Id: Mardi.cpp 17 2005-07-27 00:10:28Z degraaf $
*
* Class encapsulating a set of Trackers, monitoring them and logging values
* tracked.  
* Public API:
*   Mardi() (constructor) - create a Mardi object, initialized with 
*       default values
*   load() - parse and load a configuration file
*   checkAll() - update all trackers and log their results
*   init() - get the initial values for all trackers
*   getInterval() - get the interval between tracker updates
*   getLog() - get the tracker log stream
*   getDefaultOverCmd() - get the default command to execute when a tracked 
*       value exceeds its upper bound
*   getDefaultUnderCmd() - get the default command to execute when a tracked 
*       value exceeds its lower bound
*   getDefaultUser() - get the name of the default user under which to run 
*       commands
*   getDefaultGroup() - get the name of the default group under which to run 
*       commands
*
* This file is part of Mardi.
*
* Mardi is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* Mardi is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Mardi; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
**************************************************/

#include <string>
#include <list>
#include <exception>
#include <cstdio>
#include <ctime>
#include <cstdlib>
#include <cstring>
using namespace std;

#include <errno.h>
#include <limits.h>
#include <libxml++/libxml++.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <libgen.h>
#include "logmsg.h"
#include "Mardi.hpp"
#include "Tracker.hpp"


/*****************************
void makepath(const char* path)
Recursively creates a path, with permissions set according to the current umask
PARAMS:  const char* path - the name of the path to create
Note: aborts on error, with a message logged though the logmsg facility
*/
static void makepath(const char* path)
{
    // return on base case
    if (strlen(path) == 1 && path[0] == '.')
        return;
    if (strlen(path) == 1 && path[0] == '/')
        return;
    
    // check if the path exists
    struct stat statbuf;
    int retval;
    retval = stat(path, &statbuf);
    if (retval == -1 && errno != ENOENT) // error
    {
        logmsg(logmsg_err, "stat: %s accessing %s (%s:%i)", strerror(errno), path, __FILE__, __LINE__);
        LOGMSG_FATAL_EXIT();
        abort();
    }
    else if (retval == -1) // path doesn't exist
    {
        // recurse until we find a path that exists
        char* pathc = strdup(path);
        if (pathc == NULL)
        {
            LOGMSG_LIB(strdup);
            LOGMSG_FATAL_EXIT();
            abort();
        }
        makepath(dirname(pathc));
        free(pathc);
    
        // make the path
        retval = mkdir(path, 0777);
        if (retval == -1)
        {
            logmsg(logmsg_err, "mkdir: %s creating %s (%s:%i)", strerror(errno), path, __FILE__, __LINE__);
            LOGMSG_FATAL_EXIT();
            abort();
        }
    }
    
    return;
}

/*Mardi::BadConfigException::BadConfigException(string d) : desc(d) 
{}

Mardi::BadConfigException::~BadConfigException() throw()
{}

const char* Mardi::BadConfigException::what() const throw()
{
    return desc.c_str();
}*/

/*****************************
Mardi::Mardi()
Constructor for Mardi
*/
Mardi::Mardi() : interval(60), logname("/var/log/mardi/mardi.log"), log(NULL), defaultUser("nobody"), defaultGroup("nobody")
{}


/*****************************
Mardi::~Mardi()
Destructor for Mardi
*/
Mardi::~Mardi()
{
    if (log != NULL)
        fclose(log);
}


/*****************************
void Mardi::load(const string)
Loads an XML configuration file
PARAMS:  const string file - the name of the config file to load
Note: on error, prints a diagnostic message to stdout and exits with status 1
*/
void Mardi::load(const string file)
{
    try
    {
        // load and parse the configuration file
        xmlpp::DomParser parser;
        //parser.set_validate();
        parser.parse_file(file);
        
        if (parser) // parsing successful
        {
            // get the root node
            const xmlpp::Node* node = parser.get_document()->get_root_node();
            if (node->get_name() != "mardi")
            {
                fprintf(stderr, "Error parsing configuration file %s: Invalid root element %s\n", file.c_str(), node->get_name().c_str());
                exit(1);
            }
            const xmlpp::Element* root = dynamic_cast<const xmlpp::Element*>(node);
            
            // parse attributes
            // interval
            xmlpp::Attribute* interval_elmt = root->get_attribute(string("interval"));
            if (interval_elmt != NULL)
            {
                interval = strtoul(interval_elmt->get_value().c_str(), NULL, 0);
                if (interval == 0 || interval == UINT_MAX)
                {
                    fprintf(stderr, "Error parsing configuration file %s: Invalid interval or value out of range\n", file.c_str());
                    exit(1);
                }
            }
            
            // log file name
            xmlpp::Attribute* logfile_elmt = root->get_attribute(string("logfile"));
            if (logfile_elmt != NULL)
                logname = logfile_elmt->get_value();
            
            // defaultOverCmd
            xmlpp::Attribute* over_elmt = root->get_attribute(string("over"));
            if (over_elmt == NULL)
            {
                fprintf(stderr, "Error parsing configuration file %s: Required element \"over\" not found\n", file.c_str());
                exit(1);
            }
            defaultOverCmd = over_elmt->get_value();
            
            // defaultUnderCmd
            xmlpp::Attribute* under_elmt = root->get_attribute(string("under"));
            if (under_elmt == NULL)
            {
                fprintf(stderr, "Error parsing configuration file %s: Required element \"under\" not found\n", file.c_str());
                exit(1);
            }
            defaultUnderCmd = under_elmt->get_value();
            
            // defaultUser
            xmlpp::Attribute* user_elmt = root->get_attribute(string("user"));
            if (user_elmt != NULL)
                defaultUser = user_elmt->get_value();
            
            // defaultGroup
            xmlpp::Attribute* group_elmt = root->get_attribute(string("group"));
            if (group_elmt != NULL)
                defaultGroup = group_elmt->get_value();
            
#ifdef DEBUG
            printf("Global configuration:\n");
            printf("  interval:          %u seconds\n", interval);
            printf("  log file:          %s\n", logname.c_str());
            printf("  default over cmd:  %s\n", defaultOverCmd.c_str());
            printf("  default under cmd: %s\n", defaultUnderCmd.c_str());
            printf("  default user:      %s\n", defaultUser.c_str());
            printf("  default group:     %s\n", defaultGroup.c_str());
            printf("\n");
#endif
            
            // parse all children (of type "tracker")
            const xmlpp::Node::NodeList children = root->get_children(string("tracker"));
            if (children.size() == 0)
            {
                fprintf(stderr, "Error parsing configuration file %s: No \"tracker\" elements found\n", file.c_str());
                exit(1);
            }
            for (xmlpp::Node::NodeList::const_iterator i=children.begin(); i != children.end(); i++)
            {
                trackers.push_back(Tracker(dynamic_cast<xmlpp::Element*>(*i), file.c_str()));
            }
        }
        else // parsing failed
        {
            fprintf(stderr, "Error parsing configuration file %s\n", file.c_str());
            exit(1);
        }
    }
    catch (const exception& e)
    {
        fprintf(stderr, "Exception caught parsing %s: %s\n", file.c_str(), e.what());
        exit(1);
    }
    
    return;
}


/*****************************
void Mardi::checkAll()
Update all trackers and log their results
*/
void Mardi::checkAll()
{
    unsigned long val;
    
    // write out the time offset
    fprintf(log, "%lu", time(NULL));
    
    // update all trackers
    for (list<Tracker>::iterator i=trackers.begin(); i!=trackers.end(); i++)
    {
        val = i->check();
        fprintf(log, "\t%lu", val);
    }
    
    // end the log line
    fprintf(log, "\n");
}


/****************************
void Mardi::init()
Opens the log file and gets the initial values for all trackers
Note: on error, aborts and logs a message through the logmsg facility
*/
void Mardi::init()
{
    umask(077);
    struct stat statbuf;
    int retval;
    
    // make sure that the log file's diretory exists
    char* pathc = strdup(logname.c_str());
    makepath(dirname(pathc));
    free(pathc);
    
    // open the log file
    log = fopen(logname.c_str(), "a");
    if (log == NULL)
    {
        logmsg(logmsg_err, "Error opening log file %s: %s\n", logname.c_str(), strerror(errno));
        LOGMSG_FATAL_EXIT();
        abort();
    }
    
    // ensure that group and other don't have write permissions
    retval = fstat(fileno(log), &statbuf);
    if (retval == -1)
    {
        logmsg(logmsg_err, "fstat: %s accessing %s (%s:%i)", strerror(errno), logname.c_str(), __FILE__, __LINE__);
        LOGMSG_FATAL_EXIT();
        abort();
    }
    if (statbuf.st_mode & (S_IWGRP|S_IWOTH))
    {
        retval = fchmod(fileno(log), statbuf.st_mode & (~S_IWGRP & ~S_IWOTH));
        if (retval == -1)
        {
            logmsg(logmsg_err, "fchmod: %s on %s (%s:%i)", strerror(errno), logname.c_str(), __FILE__, __LINE__);
            LOGMSG_FATAL_EXIT();
            abort();
        }
    }
    
    // record start time
    start = time(NULL);
    
    // get the first data point for all trackers
    for (list<Tracker>::iterator i=trackers.begin(); i!=trackers.end(); i++)
        i->init();
}


/*****************************
unsigned Mardi::getInterval()
RETURNS: the number of seconds between successive tracker updates
*/
unsigned Mardi::getInterval() const
{
    return interval;
}


/*****************************
FILE* Mardi::getLog()
RETURNS: the handle to the tracker log stream
*/
FILE* Mardi::getLog() const
{
    return log;
}


/*****************************
string Mardi::getDefaultOverCmd()
RETURNS: the default command to execute when a tracked value exceeds its upper
         bound
*/
string Mardi::getDefaultOverCmd() const
{
    return defaultOverCmd;
}


/*****************************
string Mardi::getDefaultUnderCmd()
RETURNS: the default command to execute when a tracked value exceeds its lower
         bound
*/
string Mardi::getDefaultUnderCmd() const
{
    return defaultUnderCmd;
}


/*****************************
string Mardi::getDefaultUser()
RETURNS: the default user as which to run commands
*/
string Mardi::getDefaultUser() const
{
    return defaultUser;
}


/*****************************
string Mardi::getDefaultGroup()
RETURNS: the default group as which to run commands
*/
string Mardi::getDefaultGroup() const
{
    return defaultGroup;
}

