/*************************************************
* Tracker.cpp
* Copyright (c) Rennie deGraaf, 2005.  All rights reserved.
* $Id: Tracker.cpp 17 2005-07-27 00:10:28Z degraaf $
*
* Class encapsulating a single value, the data needed to track it, its upper 
* and lower bounds, and the events to be taken if those bounds are exceeded.
* Public API: 
*   Tracker() (constructor) - create a Tracker object from an XML tracker 
*       element
*   check() - get the current value, and preform appropriate actions if it has
*       exceeded its bounds
*   init() - get the initial value
*
* This file is part of Mardi.
*
* Mardi is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* Mardi is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Mardi; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
**************************************************/

#include <string>
#include <sstream>
#include <cstdlib>
#include <cstdio>
#include <cstring>
using namespace std;

#include <libxml++/libxml++.h>
#include "Tracker.hpp"
#include "Mardi.hpp"
#include "logmsg.h"


/*****************************
string to_string(unsigned long)
Converts an unsigned long to a string;
PARAMS:  unsigned long val - the value to convert
RETURNS: val, as a string in base 10
*/
static string to_string(unsigned long val)
{
    ostringstream os;
    os << val;
    return os.str();
}


/*****************************
Tracker::Tracker(const xmlpp::Element*, const char*)
Constructor for Tracker
PARAMS:  const xmlpp::Element* element - the XML "tracker" element being parsed
         const char* file - the name of the xml file being parsed (used in 
            error messages)
Note: on error, prints a diagnostic message to stdout and exits with status 1
*/
Tracker::Tracker(const xmlpp::Element* element, const char* file) : diff(false), min(0), max(0)
{
    // name
    xmlpp::Attribute* name_elmt = element->get_attribute(string("name"));
    if (name_elmt == NULL)
    {
        fprintf(stderr, "Error parsing configuration file %s: Required element \"name\" not found in tracker\n", file);
        exit(1);
    }
    name = name_elmt->get_value();

    // diff
    xmlpp::Attribute* diff_elmt = element->get_attribute(string("diff"));
    if (diff_elmt && diff_elmt->get_value() == "true")
        diff = true;
    
    // min
    xmlpp::Attribute* min_elmt = element->get_attribute(string("min"));
    if (min_elmt != NULL)
    {
        min = strtoul(min_elmt->get_value().c_str(), NULL, 0);
        if (min == ULONG_MAX && errno == ERANGE)
        {
            fprintf(stderr, "Error parsing configuration file %s: Invalid minimum or value out of range for tracker \"%s\"\n", file, name.c_str());
            exit(1);
        }
    }
    
    // max
    xmlpp::Attribute* max_elmt = element->get_attribute(string("max"));
    if (max_elmt != NULL)
    {
        max = strtoul(max_elmt->get_value().c_str(), NULL, 0);
        if (max == ULONG_MAX && errno == ERANGE)
        {
            fprintf(stderr, "Error parsing configuration file %s: Invalid maximum or value out of range for tracker \"%s\"\n", file, name.c_str());
            exit(1);
        }
    }
    
    // over
    xmlpp::Attribute* over_elmt = element->get_attribute(string("over"));
    if (over_elmt == NULL)
        overCmd = mardi.getDefaultOverCmd();
    else
        overCmd = over_elmt->get_value();
    
    // under
    xmlpp::Attribute* under_elmt = element->get_attribute(string("under"));
    if (under_elmt == NULL)
        underCmd = mardi.getDefaultUnderCmd();
    else
        underCmd = under_elmt->get_value();
    
    // user
    xmlpp::Attribute* user_elmt = element->get_attribute(string("user"));
    if (user_elmt == NULL)
        user = mardi.getDefaultUser();
    else
        user = user_elmt->get_value();
    uid = get_user_uid(user.c_str());
    if (uid == (uid_t)-1)
    {
        fprintf(stderr, "Error looking up UID for user %s: %s\n", user.c_str(), strerror(errno));
        exit(-1);
    }
    
    // group
    xmlpp::Attribute* group_elmt = element->get_attribute(string("group"));
    if (group_elmt == NULL)
        group = mardi.getDefaultGroup();
    else
        group = group_elmt->get_value();
    gid = get_group_gid(group.c_str());
    if (gid == (gid_t)-1)
    {
        fprintf(stderr, "Error looking up GID for group %s: %s\n", group.c_str(), strerror(errno));
        exit(-1);
    }
    
    // cmd
    const xmlpp::TextNode* content = element->get_child_content();
    if (content == NULL)
    {
        fprintf(stderr, "Error parsing configuration file %s: Required content not found in tracker \"%s\"\n", file, name.c_str());
        exit(1);
    }
    getCmd = content->get_content();

#ifdef DEBUG
    printf("Tracker \"%s\":\n", name.c_str());
    printf("  cmd to get value:  %s\n", getCmd.c_str());
    printf("  track differences: %s\n", (diff == true) ? "yes" : "no");
    printf("  lower bound:       %lu\n", min);
    printf("  upper bound:       %lu\n", max);
    printf("  over cmd:          %s\n", overCmd.c_str());
    printf("  under cmd:         %s\n", underCmd.c_str());
    printf("  user:              %s\n", user.c_str());
    printf("  group:             %s\n", group.c_str());
    printf("\n");
#endif
}


/*****************************
Tracker::~Tracker()
Destructor for Tracker
*/
Tracker::~Tracker()
{}


/*****************************
unsigned long Tracker::check()
Executes a Tracker's command, checks the result against its bounds, and if a 
bound is exceeded, calls the appropriate command.
RETURNS: the result of the commmand, if diff == false
         the difference from the last result, if diff == true
*/
unsigned long Tracker::check()
{
#ifdef DEBUG
    printf("executing command %s\n", getCmd.c_str());
#endif
    string result = callProgram(getCmd);
    
    // parse the result
    unsigned long current = strtoul(result.c_str(), NULL, 0);
    if (current == ULONG_MAX && errno == ERANGE)
        logmsg(logmsg_warning, "Result %s of command for tracker %s is out of range", result.c_str(), name.c_str());
    else if (current == ULONG_MAX)
        LOGMSG_LIB(strtoul);
    
    // select the value to be compared
    unsigned long val = diff ? (current-last) : current;
    
    // check against bounds
    if (min && (val < min))
        callProgram(underCmd, val);
    if (max && (val > max))
        callProgram(overCmd, val);
    
    last = current;
#ifdef DEBUG
    printf("result: %lu\n", val);
#endif
    return val;
}


/*****************************
void Tracker::init()
Executes a Tracker's command to get an initial value
*/
void Tracker::init()
{
#ifdef DEBUG
    printf("executing command %s\n", getCmd.c_str());
#endif
    string result = callProgram(getCmd);
    
    // parse the result
    last = strtoul(result.c_str(), NULL, 0);
    if (last == ULONG_MAX && errno == ERANGE)
        logmsg(logmsg_warning, "Result %s of command for tracker %s is out of range", result.c_str(), name.c_str());
    else if (last == ULONG_MAX)
        LOGMSG_LIB(strtoul);

#ifdef DEBUG
    printf("result: %lu\n", last);
#endif
}

string Tracker::callProgram(string cmd, unsigned long val)
{
    // perform translations
    string processed_cmd;
    for (unsigned i=0; i<cmd.length(); i++)
    {
        if (cmd[i] != '$')
            processed_cmd.append(1, cmd[i]);
        else if (cmd[i+1] == '$')
        {
            processed_cmd.append(1, '$');
            i++;
        }
        else if (cmd.substr(i, 5) == "$name")
        {
            processed_cmd.append(name);
            i+=4;
        }
        else if (cmd.substr(i, 4) == "$max")
        {
            processed_cmd.append(to_string(max));
            i+=3;
        }
        else if (cmd.substr(i, 4) == "$min")
        {
            processed_cmd.append(to_string(min));
            i+=3;
        }
        else if (cmd.substr(i, 4) == "$val")
        {
            processed_cmd.append(to_string(val));
            i+=3;
        }
        else if (cmd.substr(i, 5) == "$time")
        {
            time_t t = time(NULL);
            struct tm tm;
            localtime_r(&t, &tm);
            char timebuf[100];
            strftime(timebuf, 100, "%a %b %d %T %Z %Y", &tm);
            processed_cmd.append(timebuf);
            i+=4;
        }
        else
        {
            logmsg(logmsg_warning, "Unrecognized macro at position %i in command %s for tracker %s", i, cmd.c_str(), name.c_str());
        }
    }
    
    // call the program
    //printf("calling program %s\n", processed_cmd.c_str());
    char* cmdbuf = strdup(processed_cmd.c_str()); // priv_popen requires a non-const string
    if (cmdbuf == NULL)
    {
        LOGMSG_LIB(strdup);
        LOGMSG_FATAL_EXIT();
        abort();
    }
    popen_handle_t* program = priv_popen(cmdbuf, uid, gid);
    if (program == NULL)
        return string(""); // logging already taken care of
    free(cmdbuf);
    
    // read the result
    string result;
    char buf[100];
    while (!feof(program->read_fd))
    {
        fgets(buf, 99, program->read_fd);
        result.append(buf);
    }
    //printf("result: %s\n", result.c_str());
    
    // get the exit code
    int retval = priv_pclose(program);
    if (retval == 0)
        return result;
    else if (retval == -1)
        return string(""); // logging already taken care of
    else if (retval == -2)
    {
        logmsg(logmsg_warning, "Command %s exited abnormally with output \"%s\"", processed_cmd.c_str(), result.c_str());
        return string("");
    }
    else
    {
        logmsg(logmsg_warning, "Command %s exited with error code %i and output \"%s\"", processed_cmd.c_str(), retval, result.c_str());
        return string("");
    }
}

