/*************************************************
* mardi_main.cpp
* Copyright (c) Rennie deGraaf, 2005.  All rights reserved.
* $Id: mardi_main.cpp 15 2005-07-26 07:02:21Z degraaf $
*
* Main module for Mardi - a system to track integer values from the system,
* log them in a way that can be graphed with GNUPlot, and execute commands if
* any of those values exceed set upper and lower bounds.
*
* This file is part of Mardi.
*
* Mardi is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* Mardi is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Mardi; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
**************************************************/

#define _MAIN_

#include <cstdlib>
#include <cstdio>
#include <cctype>
using namespace std;

#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "Mardi.hpp"
#include "logmsg.h"
#include "spc_sanitize.h"
#include "daemonize.h"

/* global config object */
Mardi mardi;

#define QUOTE(X) XQUOTE(X)
#define XQUOTE(X) #X


/*****************************
void version()
Prints version info
*/
void version()
{
    printf("Mardi %s\n", QUOTE(VERSION));
    printf("Copyright (c) Rennie deGraaf, 2005.  All rights reserved.\n");
}

/*****************************
void printhelp()
Print a help message
*/
void printhelp()
{
    printf("Usage: mardi [-c <config file>] [-D] [-h] [-v]\n");
    printf("where -c <config file> - use the specified configuration file\n");
    printf("      -D - run the program as a daemon\n");
    printf("      -h - print this message\n");
    printf("      -v - print version information\n");
    return;
}


/*****************************
void alarm_handler(int)
Handler function for SIGALRM - updates all trackers
PARAMS:  int sig - the signal being caught
*/
static void alarm_handler(int)
{
    alarm(mardi.getInterval());
    //printf("we get signal\n");
    mardi.checkAll();
    return;
}


/*****************************
void sighup_handler(int)
Handler function for SIGHUP - flushes the log to disk
PARAMS:  int sig - the signal being caught
*/
static void sighup_handler(int)
{
    fflush(mardi.getLog());
}


/*****************************
void sigint_handler(int)
Handler function for SIGINT - exits normally
PARAMS:  int sig - the signal being caught
*/
static void sigint_handler(int)
{
    logmsg(logmsg_notice, "Shutting down normally");
    logmsg_close();
    exit(0);
}

/*****************************
int main(const int, const char**)
Main routine.  If you don't know what this does, you should learn C++
*/
int main(int argc, char** argv)
{
    sighandler_t oldsig;
    bool daemon = false;
    string cfg_file = "/etc/mardi.conf";
    int c;

    // initialize the logmsg facility
    logmsg_open(logmsg_stderr, 0, argv[0]);
    
    // sanitize the system
    spc_sanitize_environment(0, NULL);
    spc_sanitize_files();

    /* parse command-line args */
    while ((c = getopt(argc, argv, "c:Dhv")) != -1)
    {
        switch (c)
        {
            case 'c':
                cfg_file = optarg;
                break;
            case 'D':
                daemon = true;
                break;
            case 'h':
                version();
                printhelp();
                exit(0);
            case 'v':
                version();
                exit(0);
            case '?':
                if (isprint(optopt))
                    fprintf(stderr, "Unrecognized option -%c\n", optopt);
                else
                    fprintf(stderr, "Unrecognized option character 0x%x\n", optopt);
                exit(1);
            case ':':
                fprintf(stderr, "Argument required to option -%c\n", optopt);
                exit(1);
            default:
                fprintf(stderr, "getopt() returned an unexpected value\n");
                exit(1);
        }
    }
    
    /* check permissions on the log file */
    struct stat statbuf;
    int retval = stat(cfg_file.c_str(), &statbuf);
    if (retval == -1)
    {
        fprintf(stderr, "Error opening configuration file %s: %s\n", cfg_file.c_str(), strerror(errno));
        exit(1);
    }
    // make sure no one else can write to the config file
    if (statbuf.st_mode & (S_IWGRP|S_IWOTH))
    {
        fprintf(stderr, "Inappropriate permissions on configuration file %s\n", cfg_file.c_str());
        exit(1);
    }
    
    /* load configuration */
    //try
    //{
    mardi.load(cfg_file);
    //} 
    //catch (const Mardi::BadConfigException& e)
    //{
    //    fprintf(stderr, "Error loading configuration file %s: %s\n", cfg_file.c_str(), e.what());
    //    return 1;
    //}
    
    /* initialize */
    mardi.init();

    /* we've finished initializing; time to daemonize */
    if (daemon)
    {
        daemonize("/dev/null", "/tmp/out", "/dev/null");
        
        // daemonize closes logmsg; re-open it
        logmsg_open(logmsg_syslog, 0, argv[0]);
    }
    logmsg(logmsg_notice, "Starting up...");
    
    // register signal handlers
    // we can't do this until after we've daemonized
    alarm(mardi.getInterval());
    oldsig = signal(SIGALRM, alarm_handler);
    if (oldsig == SIG_ERR)
    {
        logmsg(logmsg_err, "Error registering signal handler for SIGALRM (%s:%i)", __FILE__, __LINE__);
        LOGMSG_FATAL_EXIT();
        abort();
    }
    oldsig = signal(SIGHUP, sighup_handler);
    if (oldsig == SIG_ERR)
    {
        logmsg(logmsg_err, "Error registering signal handler for SIGHUP (%s:%i)", __FILE__, __LINE__);
        LOGMSG_FATAL_EXIT();
        abort();
    }
    oldsig = signal(SIGINT, sigint_handler);
    if (oldsig == SIG_ERR)
    {
        logmsg(logmsg_err, "Error registering signal handler for SIGINT (%s:%i)", __FILE__, __LINE__);
        LOGMSG_FATAL_EXIT();
        abort();
    }
    
    /* block indefinitely and let the signal handlers work their magic */
    while (1)
    {
        pause();
    }
    
    return 0;
}
