/*************************************************
* priv_popen.c
* Copyright (c) 2003-2005, Rennie deGraaf, Matt Messier, John Viega.  
* All rights reserved.
* $Id: priv_popen.c 14 2005-07-26 02:00:59Z degraaf $

* A full-duplex variation on popen() with privilege separation.
* Executes a given command in a restricted Bourne shell (/bin/sh -r),
* connecting its stdin and stdout to FILE objects.  If run with superuser
* privileges, it sets its UID and GID to supplied values first.
* 
* priv_popen() - Starts a program and opens pipes to its stdin and stdout
* priv_pclose() - Closes the pipes to a program and waits for it to close
*
* Based on code from "Secure Programming Cookbook for C and C++" by 
* Matt Messier and John Viega.
*
* This file is part of the libwheel project.
*
* libwheel is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* libwheel is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with libwheel; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
**************************************************/


#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/wait.h>
#include <assert.h>
#include "drop_priv.h"
#include "priv_popen.h"
#include "logmsg.h"
#include "spc_sanitize.h"


/*****************************
popen_handle_t* priv_popen(char*, const uid_t, const gid_t
Runs a command in a restricted Bourne shell with a given UID and GID, attaching 
file streams to its stdin and stdout
PARAMS:  char* command - the shell command to run
         const uid_t newuid - the UID to run the command under
         const gid_t newgid - the GID to run the command under
RETURNS: a handle to the new process on success
         NULL on error, sends error messages through the logmsg facility
Note: - assumes that the current environemt is safe
    - if the child fails before launching the command, it aborts
*/
/* assumes that the environment is safe */
popen_handle_t* priv_popen(char *command, const uid_t newuid, const gid_t newgid) 
{
    int stdin_pipe[2], stdout_pipe[2];
    popen_handle_t* p;
    char* args[] = {"/bin/sh", "-r", "-c", command, NULL};

    p = malloc(sizeof(popen_handle_t));
    if (p == NULL)
    {
        LOGMSG_LIB(pipe);
        return NULL;
    }
    p->read_fd = p->write_fd = 0;
    p->child_pid = -1;

    if (pipe(stdin_pipe) == -1) 
    {
        LOGMSG_LIB(pipe);
        free(p);
        return NULL;
    }
    if (pipe(stdout_pipe) == -1) 
    {
        LOGMSG_LIB(pipe);
        close(stdin_pipe[1]);
        close(stdin_pipe[0]);
        free(p);
        return NULL;
    }

    if (!(p->read_fd = fdopen(stdout_pipe[0], "r"))) 
    {
        LOGMSG_LIB(fdopen);
        close(stdout_pipe[1]);
        close(stdout_pipe[0]);
        close(stdin_pipe[1]);
        close(stdin_pipe[0]);
        free(p);
        return NULL;
    }
    if (!(p->write_fd = fdopen(stdin_pipe[1], "w"))) 
    {
        LOGMSG_LIB(fdopen);
        fclose(p->read_fd);
        close(stdout_pipe[1]);
        close(stdin_pipe[1]);
        close(stdin_pipe[0]);
        free(p);
        return NULL;
    }

    if ((p->child_pid = fork()) == -1) 
    {
        LOGMSG_LIB(fork);
        fclose(p->write_fd);
        fclose(p->read_fd);
        close(stdout_pipe[1]);
        close(stdin_pipe[0]);
        free(p);
        return NULL;
    }

    if (!p->child_pid) 
    {
        /* this is the child process */
        close(stdout_pipe[0]);
        close(stdin_pipe[1]);
        if (stdin_pipe[0] != 0) {
            close(0);
            if (dup2(stdin_pipe[0], 0) == -1)
            {
                LOGMSG_LIB(dup2);
                LOGMSG_FATAL_EXIT();
                abort();
            }
            close(stdin_pipe[0]);
        }
        if (stdout_pipe[1] != 1) {
            /* note: logmsg may not work reliably after this point */
            close(1);
            if (dup2(stdout_pipe[1], 1) == -1)
            {
                LOGMSG_LIB(dup2);
                LOGMSG_FATAL_EXIT();
                abort();
            }
            close(stdout_pipe[1]);
        }
    
        /* close open files */
        spc_sanitize_files();
    
        /* drop privileges */
        if (geteuid() == 0)
            drop_priv(newuid, newgid);
    
        /* NOTE: this uses the shell */
        execv(args[0], args);
    
        /* this only happens if execv fails */
        LOGMSG_LIB(execv);
        LOGMSG_FATAL_EXIT();
        abort();
    }

    close(stdout_pipe[1]);
    close(stdin_pipe[0]);
    return p;
}


/*****************************
int priv_pclose(popen_handle_t*)
Waits for a process opened with priv_popen to exit, and closes the file streams
associated with it
PARAMS:  popen_handle_t* p - the handle to the process
RETURNS: the process' exit code, if it exited normally
         -1 on failure (with a message logged to the logmsg facility)
         -2 if the process exited abnormally (including if priv_popen failed
            before launching its commnand)
*/
int priv_pclose(popen_handle_t* p) 
{
    int   status;
    pid_t pid = -1;
  
    assert(p != NULL);

    if (p->child_pid != -1) 
    {
        do 
        {
            pid = waitpid(p->child_pid, &status, 0);
        } while (pid == -1 && errno == EINTR);
        if (pid == -1)
        {
            LOGMSG_LIB(waitpid);
        }
    }
    if (p->read_fd) 
        fclose(p->read_fd);
    if (p->write_fd) 
        fclose(p->write_fd);
    free(p);
    if (pid == -1)
        return -1;
    else if (WIFEXITED(status)) 
        return WEXITSTATUS(status);
    else 
        return -2;
}
